# -*- coding: utf-8 -*-
#
# Copyright 2012 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
    Bulbs is a Python persistence framework for graph databases.
    
    Bulbs is a Python client library for Neo4j Server and Rexster, 
    Rexster is a REST server that provides access to any 
    Blueprints-enabled graph database, including Neo4j, OrientDB, DEX, 
    TinkerGraph and OpenRDF.

    Home Page: http://bulbflow.com
 
    Copyright (c) 2012, James Thornton (http://jamesthornton.com)
    License: BSD (see LICENSE for details)
"""
__version__ = '0.3'
__author__ = 'James Thornton'
__license__ = 'BSD'





